import subprocess
import json
import copy
from . import configuration as cfg
from . import identity
from httpx import Client, HTTPStatusError
from .resource_abc import Resource, Ref
from typing import Optional, Dict, Any, Union, Literal, List
from datetime import datetime
import time
from pydantic import BaseModel, computed_field, field_serializer, Field


class ImageRef(BaseModel, Ref):
    """Reference an existing image
    Example
    -------
    >>> from fbnconfig.scheduler import ImageRef
    >>> image_ref = ImageRef(dest_name="myimage", dest_tag="0.1.1")
    """
    id: str
    dest_name: str
    dest_tag: str

    def attach(self, client):
        # we only need to check it exists
        client.get(f"/scheduler2/api/images/{self.dest_name}:{self.dest_tag}")


class ImageResource(BaseModel, Resource):
    """Define an image

    Example
    -------
    >>> from fbnconfig.scheduler import ImageResource
    >>> image = ImageResource(
            id="img1",
            source_image="docker.io/alpine:3.16.7",
            dest_name="myimage",
            dest_tag="3.16.7")
    """
    id: str
    source_image: str
    dest_name: str
    dest_tag: str

    def read(self, client, old_state):
        pass

    def create(self, client):
        env_name = client.base_url.host.split(".")[0]

        downstream_image = f"{client.base_url.host}/{env_name}/{self.dest_name}:{self.dest_tag}"

        # pull from the upstream and apply the downstream tag
        pull_commands = self._pull_commands(downstream_image)

        for cmd in pull_commands:
            subprocess.run(cmd).check_returncode()
        # get local and downstream digests to see if they match
        local_digest = self._getLocalDigest(downstream_image)

        if local_digest is not None:
            downstream_images = self._getRemoteTags(client, self.dest_name, local_digest)
            matching_tag = next((tag for tag in downstream_images if tag["name"] == self.dest_tag), None)
            if matching_tag is not None:  # this image/tag combo already exists
                print("image already exists in the remote with the same digest")
                return {
                    "id": self.id,
                    "source_image": self.source_image,
                    "dest_name": self.dest_name,
                    "dest_tag": self.dest_tag,
                }
        # create a unique tag so scheduler will give us auth info and not complain about
        # reusing an existing one
        body = {"imageName": f"{self.dest_name}:{time.time()}"}
        upload = client.post("/scheduler2/api/images", json=body)
        auth = upload.json()["dockerLoginCommand"].split(" ")

        # push the tag we actually want to have using the auth from scheduler
        push_commands = self._push_commands(downstream_image, auth)

        for cmd in push_commands:
            subprocess.run(cmd).check_returncode()

        return {
            "id": self.id,
            "source_image": self.source_image,
            "dest_name": self.dest_name,
            "dest_tag": self.dest_tag,
        }

    def _getRemoteTags(self, client, dest_name, local_digest):
        repo = client.get(f"/scheduler2/api/images/repository/{dest_name}")
        # todo: this could be paged
        return next(
            (img["tags"] for img in repo.json()["values"] if img["digest"] == local_digest),
            [],
        )

    def _getLocalDigest(self, tag):
        inspect = subprocess.run(
            ["docker", "inspect", "--format", "{{json .RepoDigests}}", tag],
            capture_output=True,
            text=True,
        )
        inspect.check_returncode()
        # inspect returns:
        #   [alpine@sha256:a8cb.. fbn-qa.lusid.com/fbn-qa/beany@sha256:b797...]
        repo = tag.split(":")[0]
        return next(
            (
                digest.split("@")[1]
                for digest in json.loads(inspect.stdout)
                if digest.split("@")[0] == repo
            ),
            None,
        )

    def _push_commands(self, tag, auth):
        return [
            ["docker", "login", "-u", auth[3], "--password", auth[5], auth[6]],
            ["docker", "push", tag],
        ]

    def _pull_commands(self, tag):
        return [
            ["docker", "pull", "--platform", "linux/amd64", self.source_image],
            ["docker", "tag", self.source_image, tag],
        ]

    def update(self, client, old_state):
        if (
            self.source_image == old_state.source_image
            and self.dest_name == old_state.dest_name
            and self.dest_tag == old_state.dest_tag
        ):
            return None
        return self.create(client)

    @staticmethod
    def delete(client, old_state):
        # no delete for images
        pass

    def deps(self):
        return []

class CommandlineArg(BaseModel):
    """Define a commandline ArgumentDefinition for a JobResource"""
    dataType: Literal["String", "Int"]
    required: Optional[bool] = False
    description: str
    order: int
    passedAs: Literal["CommandLine"] = "CommandLine"
    defaultValue: None|str = None

class EnvironmentArg(BaseModel):
    """Define a environment ArgumentDefinition for a JobResource"""
    dataType: Literal["String", "Int", "SecureString", "Configuration"]
    required: Optional[bool] = False
    description: str
    order: int
    passedAs: Literal["EnvironmentVariable"] = "EnvironmentVariable"
    defaultValue: None|str|cfg.ItemRef|cfg.ItemResource = None

    @field_serializer("defaultValue", when_used="always")
    def value_ser(self, value) -> str|None:
        if value is None:
            return None
        if isinstance(value, (cfg.ItemRef, cfg.ItemResource)):
            return value.ref
        return value

class JobRef(BaseModel, Ref):
    """Reference an existing scheduler job"""
    id: str = Field(exclude=True)
    scope: str
    code: str

    def attach(self, client):
        # just check it exists
        search = client.request(
            "get",
            "/scheduler2/api/jobs",
            params={"filter": f"jobId.scope eq '{self.scope}' and jobId.code eq '{self.code}'"},
        )

        current = next((job for job in search.json()["values"]), None)
        if current is None:
            raise RuntimeError(
                f"Failed to attach JobRef to job with scope={self.scope} and code={self.code}"
            )


class JobState(BaseModel):
    id: str
    scope: str
    code: str


class JobResource(BaseModel, Resource):
    """Manage a JobDefinition"""
    id: str = Field(exclude=True)
    scope: str = Field(exclude=True, init=True)
    code: str = Field(exclude=True, init=True)
    image: Union[ImageResource, ImageRef] = Field(exclude=True, init=True)
    name: str
    author: Optional[str] = None
    dateCreated: Optional[datetime] = None
    description: str
    ttl: Optional[int] = None
    minCpu: Optional[str] = None
    maxCpu: Optional[str] = None
    minMemory: Optional[str] = None
    maxMemory: Optional[str] = None
    argumentDefinitions: Dict[str, CommandlineArg|EnvironmentArg] = {}
    commandLineArgumentSeparator: Optional[str] = None
    remote: Dict[str, Any] = Field({}, exclude=True, init=False)

    @computed_field
    def jobId(self) -> Dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    @computed_field
    def imageName(self) -> str:
        return self.image.dest_name

    @computed_field
    def imageTag(self) -> str:
        return self.image.dest_tag

    def read(self, client: Client, old_state):
        search = client.get(
            "/scheduler2/api/jobs",
            params={
                "filter": f"jobId.scope eq '{old_state.scope}' and jobId.code eq '{old_state.code}'"
            },
        )
        current = next(job for job in search.json()["values"])
        # normalise the current values
        current.pop("jobId")
        current.pop("requiredResources")
        current["imageName"] = current["dockerImage"].split(":")[
            0
        ]  # we get dockerImage but have to send separately :(
        current["imageTag"] = current["dockerImage"].split(":")[1]
        current.pop("dockerImage")
        self.remote = current

    def create(self, client):
        desired = self.model_dump(mode="json", exclude_none=True)
        client.post("/scheduler2/api/jobs", json=desired)
        newState = JobState(
            id=self.id,
            scope=self.scope,
            code=self.code
        )
        return newState.model_dump()

    def update(self, client: Client, old_state):
        if self.scope != old_state.scope or self.code != old_state.code:
            raise (RuntimeError("Cannot change identifier on job. Create a new one"))
        self.read(client, old_state)
        remote = copy.deepcopy(self.remote)
        desired = self.model_dump(mode="json", exclude_none=True, exclude={"jobId"})
        effective = remote | desired
        remoteArgs = remote["argumentDefinitions"]
        effectiveArgs = {
            argKey: remoteArgs.get(argKey, {}) | argValue
            for argKey, argValue in desired["argumentDefinitions"].items()
        }
        effective["argumentDefinitions"] = effectiveArgs
        if effective == remote:
            return None
        client.put(f"/scheduler2/api/jobs/{self.scope}/{self.code}", json=desired)
        new_state = JobState(
            id=self.id,
            scope=self.scope,
            code=self.code
        )
        return new_state.model_dump()

    @staticmethod
    def delete(client: Client, old_state):
        client.delete(f"/scheduler2/api/jobs/{old_state.scope}/{old_state.code}")

    def cleanup(self, client, oldState):
        if self.scope != oldState.scope or self.code != oldState.code:
            return self.delete(client, oldState)
        return None

    def deps(self):
        config_args: List[cfg.ItemRef|cfg.ItemResource] = [
            arg.defaultValue for arg in self.argumentDefinitions.values()
            if isinstance(arg.defaultValue, (cfg.ItemRef, cfg.ItemResource))
        ]
        return [self.image] + config_args


class ScheduleState(BaseModel):
    id: str
    scope: str
    code: str
    argKeys: List[str]


class ScheduleResource(BaseModel, Resource):
    """Manage a ScheduleDefinition"""
    id: str = Field(exclude=True)
    name: str
    scope: str = Field(exclude=True, init=True)
    code: str = Field(exclude=True, init=True)
    expression: str = Field(exclude=True, init=True)
    timezone: str = Field(exclude=True, init=True)
    job: JobResource | JobRef = Field(serialization_alias="jobId")
    description: str
    author: Optional[str] = None
    owner: Optional[str] = None
    arguments: None | Dict[str, str|cfg.ItemResource|cfg.ItemRef] = None
    enabled: Optional[bool] = None
    useAsAuth: identity.UserResource | identity.UserRef | None = None
    _remote: Dict[str, Any] = {}

    @field_serializer("useAsAuth", when_used="always")
    def serialize_auth(self, user: identity.UserResource | identity.UserRef | None = None):
        return user.userId if user else None

    @field_serializer("arguments", when_used="always")
    def serialize_args(self, args) -> Dict[str, str]:
        return {
            k: v.ref if isinstance(v, (cfg.ItemRef, cfg.ItemResource)) else v
            for k,v in args.items()
        }

    @field_serializer("job", when_used="always",)
    def serialize_job(self, job: JobResource | JobRef):
        return {"scope": job.scope, "code": job.code}

    @computed_field
    def scheduleId(self) -> Dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    @computed_field
    def trigger(self) -> Dict[str, Dict[str,str]]:
        return {
            "timeTrigger": {
                "expression": self.expression,
                "timeZone": self.timezone,
            }
        }

    def read(self, client, old_state):
        get = client.get(f"/scheduler2/api/schedules/{old_state.scope}/{old_state.code}")
        current = get.json()
        # note scheduleId for create scheduleIdentifier for read :(
        current.pop("scheduleIdentifier")
        self._remote = current

    def create(self, client):
        body = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        client.post("/scheduler2/api/schedules", json=body)
        arg_keys = list(self.arguments.keys()) if self.arguments else []
        return ScheduleState(id=self.id, scope=self.scope, code=self.code, argKeys=arg_keys).model_dump()

    def update(self, client, old_state):
        if self.scope != old_state.scope or self.code != old_state.code:
            from_id = f"{old_state.scope}/{old_state.code}"
            to_id = f"{self.scope}/{self.code}"
            raise (RuntimeError(f"Cannot change schedule identifier. From: '{from_id}', to: '{to_id}'"))
        self.read(client, old_state)
        remote = copy.deepcopy(self._remote)
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True, exclude={"scheduleId"})
        effective = remote | desired
        # On read the schedule will contain schedule args and args from the job.
        # Only update if the user specified args are different to the remote
        effective["arguments"] = remote.get("arguments", {}) | desired.get("arguments", {})
        arg_keys = list(self.arguments.keys()) if self.arguments else []
        if effective == remote and arg_keys == getattr(old_state, "argKeys", []):
            return None
        client.put(f"/scheduler2/api/schedules/{self.scope}/{self.code}", json=desired)
        return ScheduleState(id=self.id, scope=self.scope, code=self.code, argKeys=arg_keys).model_dump()

    @staticmethod
    def delete(client, old_state):
        try:
            client.delete(f"/scheduler2/api/schedules/{old_state.scope}/{old_state.code}")
        except HTTPStatusError as ex:
            content = ex.response.json()
            if (
                ex.response.status_code == 404
                and content.get("name", None) == "ValidationError"
                and content.get("title", None) == "Schedule could not be found"
            ):
                pass
            else:
                raise ex

    def cleanup(self, client, old_state):
        if self.scope != old_state.scope or self.code != old_state.code:
            return self.delete(client, old_state)
        return None

    def deps(self):
        config_args: List[cfg.ItemRef|cfg.ItemResource] = [
            value for value in self.arguments.values()
            if isinstance(value, (cfg.ItemRef, cfg.ItemResource))
        ] if self.arguments else []
        return [self.job] + config_args



from types import SimpleNamespace
import httpx
from abc import ABC, abstractmethod
from typing import Dict, Any, Union, List, Optional


class Ref(ABC):
    @abstractmethod
    def attach(self, client: httpx.Client) -> None:
        pass

    def deps(self):
        return []


class Resource(ABC):
    @abstractmethod
    def read(self, client: httpx.Client, old_state: SimpleNamespace) -> None:
        pass

    @abstractmethod
    def create(self, client) -> Optional[Dict[str, Any]]:
        pass

    @abstractmethod
    def update(self, client, old_state) -> Union[None, Dict[str, Any]]:
        pass

    @staticmethod
    @abstractmethod
    def delete(client, old_state) -> None:
        pass

    @abstractmethod
    def deps(self) -> List[Any]:
        pass



import json
from collections import defaultdict
from types import SimpleNamespace
from typing import Optional, Dict, List, NamedTuple
from datetime import datetime

import httpx

LogLine = NamedTuple("LogLine", [
    ("entry_id", str),
    ("deployment", str),
    ("resource_id", str),
    ("state", str),
    ("dependencies", List[str]),
    ("resource_type", str),
    ("as_at", str)
])
DepList = Dict[str, List[str]]
Index = Dict[str, LogLine]


def record(client, deployment_id, resource, state):
    res = client.request(
        "post",
        "/api/api/customentities/~deployment",
        json={
            "displayName": resource.id,
            "description": "...",
            "identifiers": [
                {
                    "identifierScope": "deployment",
                    "identifierType": "resource",
                    "identifierValue": f"{deployment_id}_{resource.id}".replace("/", "_"),
                }
            ],
            "fields": [
                {
                    "name": "deployment",
                    "value": deployment_id,
                },
                {
                    "name": "resourceType",
                    "value": type(resource).__name__,
                },
                {"name": "dependencies", "value": [r.id for r in resource.deps()]},
                {
                    "name": "resource",
                    "value": resource.id,
                },
                {"name": "state", "value": json.dumps(state)},
            ],
        },
    )
    print("Logged changes to:", f"{deployment_id}_{resource.id}")
    return res.json()

def fetch_log_entries(client: httpx.Client, params: Dict):
    page = None
    while True:
        pp = params | {"page": page} if page is not None else params
        res = client.get( "/api/api/customentities/~deployment", params=pp).json()
        for item in res["values"]:
            yield item
        if res.get("nextPage") is not None:
            page = res["nextPage"]
        else:
            break

def entry_to_logline(entry: Dict):
    indexed = {
        "identifiers": {
            i["identifierType"]: i["identifierValue"]
            for i in entry["identifiers"]
        },
        "fields": {
            f["name"]: f["value"]
            for f in entry["fields"]
        },
        "as_at": entry["version"]["asAtModified"],
    }
    parsed = {
        "entry_id": indexed["identifiers"]["resource"],
        "deployment": indexed["fields"]["deployment"],
        "resource_id": indexed["fields"]["resource"],
        "state": SimpleNamespace(**json.loads(indexed["fields"]["state"])),
        "dependencies": indexed["fields"]["dependencies"],
        "resource_type": indexed["fields"]["resourceType"],
        "as_at": indexed["as_at"],
    }
    return LogLine(**parsed)


def list_resources_for_deployment(client: httpx.Client, deployment_id: str) -> List[LogLine]:
    lines = [
        entry_to_logline(e)
        for e in fetch_log_entries(client, {"filter": f"fields[deployment] eq '{deployment_id}'"})
    ]
    lines.sort(key=lambda line: datetime.fromisoformat(line.as_at), reverse=True)
    return lines


def list_deployments(client: httpx.Client) -> List[str]:
    log_lines = [
        entry_to_logline(e)
        for e in fetch_log_entries(client, {})
    ]
    log_lines.sort(key=lambda line: datetime.fromisoformat(line.as_at), reverse=True)
    seen = []
    # get sorted set of deployments
    for line in log_lines:
        if line.deployment not in seen:
            seen.append(line.deployment)
    return seen


def get_resource(client, deployment_id, resource_id) -> List[LogLine]:
    filter = f"fields[deployment] eq '{deployment_id}' and fields[resource] eq '{resource_id}'"
    log_lines = [
        entry_to_logline(e)
        for e in fetch_log_entries(client, {"filter": filter})
    ]
    return log_lines


# index of dependencies (key resource needs values to exist)
def entry_deps(log: List[LogLine]) -> DepList:
    return {e.resource_id: e.dependencies for e in log}


# entries indexed by resource id
def entry_index(log: List[LogLine]) -> Index:
    return {e.resource_id: e for e in log}


# index of dependants (value resources need key resource to exist)
def entry_rdeps(log: List[LogLine]) -> DepList:
    reverse_deps = defaultdict(list)
    for e in log:
        for d in e.dependencies:
            reverse_deps[d].append(e.resource_id)
    return reverse_deps


def print_tree(index: Dict[str, LogLine], deps: DepList, start_at: str, marker: str) -> None:
    leader = "    "

    def helper(resource_id, depth):
        if depth == 0:
            print(resource_id,
                  f"({index[resource_id].resource_type})"
                  if index.get(resource_id, None) else "Unknown Type")
        else:
            print(leader * depth, marker if depth > 0 else "", resource_id,
                  f"({index[resource_id].resource_type})"
                  if index.get(resource_id, None) else "Unknown Type")
        for d in deps.get(resource_id, {}):
            if d:
                helper(d, depth + 1)
            else:
                helper(resource_id, 0)

    helper(start_at, 0)


def get_dependencies_map(client, deployment_id) -> tuple[Index, DepList]:
    resources: List[LogLine] = list_resources_for_deployment(client, deployment_id)

    return entry_index(resources), entry_deps(resources)


def get_dependents_map(client, deployment_id) -> tuple[Index, DepList]:
    resources: List[LogLine] = list_resources_for_deployment(client, deployment_id)

    return entry_index(resources), entry_rdeps(resources)


def remove(client, deployment_id, resource_id):
    res = client.request(
        "delete",
        f"/api/api/customentities/~deployment/resource/{deployment_id}_{resource_id}",
        params={"identifierScope": "deployment"},
    )
    print("Removed log for:", f"{deployment_id}_{resource_id}")
    return res.json()


def format_log_line(line: LogLine):
    return f"id: {line.resource_id} | type: {line.resource_type} | {line.dependencies}"


def identifier_create(client) -> Optional[Dict]:
    try:
        res = client.request(
            "post",
            "/api/api/propertydefinitions",
            json={
                "domain": "CustomEntity",
                "scope": "deployment",
                "code": "resource",
                "displayName": "Deployment Resource",
                "dataTypeId": {"scope": "system", "code": "string"},
                "lifeTime": "Perpetual",
                "constraintStyle": "Identifier",
                "description": "To identify resources in a deployment",
            },
        )
        print("Created identifier", res.json())
        return res.json()
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 400 and e.response.json().get("name") == "PropertyAlreadyExists":
            print("   Identifier already exists")
            return None
        raise e


entity_definition = {
    "displayName": "Deploy Resource",
    "description": "Deploy Resource v2.1",
    "fieldSchema": [
        {
            "name": "dependencies",
            "lifetime": "Perpetual",
            "type": "String",
            "collectionType": "Array",
            "required": True,
            "description": "IDs of resource this one depends on",
        },
        {
            "name": "deployment",
            "lifetime": "Perpetual",
            "type": "String",
            "required": True,
            "description": "the deployment this resource is in",
        },
        {
            "name": "resource",
            "lifetime": "Perpetual",
            "type": "String",
            "required": True,
            "description": "ID of the resource within the deployment",
        },
        {
            "name": "resourceType",
            "lifetime": "Perpetual",
            "type": "String",
            "required": True,
            "description": "Class that manages this resource",
        },
        {
            "name": "state",
            "lifetime": "Perpetual",
            "type": "String",
            "required": True,
            "description": "Current state resource within the deployment",
        },
    ],
}


def entity_def_get(client) -> Optional[Dict]:
    try:
        res = client.get("/api/api/customentitytypes/~deployment")
        return res.json()
    except httpx.HTTPStatusError as e:
        if (
                e.response.status_code == 404
                and e.response.json().get("name") == "CustomEntityDefinitionNotFound"
        ):
            return None
        raise e


def entity_def_create(client) -> Dict:
    body = {"entityTypeName": "deployment"} | entity_definition
    res = client.request("post", "/api/api/customentitytypes", json=body)
    return res.json()


def entity_def_update(client) -> Dict:
    res = client.request("put", "/api/api/customentitytypes/~deployment", json=entity_definition)
    return res.json()


def entity_def_upsert(client):
    existing_entity_def = entity_def_get(client)

    if existing_entity_def is not None:
        existing_entity_def.pop("href", None)
        existing_entity_def.pop("entityType", None)
        existing_entity_def.pop("version", None)
        existing_entity_def.pop("entityTypeName", None)

        if existing_entity_def == entity_definition:
            print("   Entity already exists")
        else:
            update = entity_def_update(client)
            print(update)
    else:
        create = entity_def_create(client)
        print("created", create)


def setup(client):
    identifier_create(client)
    entity_def_upsert(client)



import httpx
import os
import json
from types import SimpleNamespace
from typing import NamedTuple, List, Union
from . import log
from .resource_abc import Resource, Ref

# import all resources here, so they are available in globals()
from .drive import FolderResource, FileResource  # noqa: F401
from .scheduler import ImageResource, JobResource, ScheduleResource  # noqa: F401
from .access import PolicyResource, RoleResource  # noqa: F401
from .identity import IdentityRoleResource, UserResource, RoleAssignment  # noqa: F401
from .configuration import SetResource, ItemResource # noqa: F401
from .property import DefinitionResource # noqa: F401

Deployment = NamedTuple("Deployment", [("id", str), ("resources", List[Union[Resource, Ref]])])
deployments = []


def deployment(id, resources):
    d = Deployment(id, resources)
    deployments.append(d)


def recurse_resources(resources):
    seen = set()
    ordered = []

    def helper(items):
        for res in items:
            deps = res.deps()
            if len(deps) > 0:
                helper(deps)
            if res.id not in seen:
                seen.add(res.id)
                ordered.append(res)

    helper(resources)
    return ordered


def sort_deployed(entries):
    seen = set()
    ordered = []

    def helper(items):
        for e in items:
            dep_ids = e.dependencies
            if len(dep_ids) > 0:
                deps = [e for e in entries if e.resource_id in dep_ids]
                helper(deps)
            if e.resource_id not in seen:
                seen.add(e.resource_id)
                ordered.append(e)

    helper(entries)
    return ordered


# Throw an exception whenever the response code is 4xx or 5xx
def response_hook(response: httpx.Response) -> None:
    if response.is_error:
        response.read()
        response.raise_for_status()


def create_client(lusid_url: str, token: str) -> httpx.Client:
    if token is None or lusid_url is None:
        raise (
            RuntimeError(
                "Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set"
            )
        )

    return httpx.Client(
        base_url=lusid_url,
        timeout=20,
        headers={
            "authorization": "Bearer " + os.getenv("FBN_ACCESS_TOKEN", default=""),
            "accept": "application/json",
            "content-type": "application/json",
        },
        event_hooks={"response": [response_hook]},
    )


def run(client: httpx.Client, deployment):
    resources = deployment.resources
    deploy_state = log.list_resources_for_deployment(client, deployment.id)
    all_resources = recurse_resources(resources)
    for resource in all_resources:
        if isinstance(resource, Ref):
            resource.attach(client)
            print("attach", type(resource).__name__)
            continue
        match = [e for e in deploy_state if e.resource_id == resource.id]
        if len(match) == 0:
            print("deploy: create", type(resource).__name__, resource.id)
            state = resource.create(client)
        else:
            print("deploy: update", type(resource).__name__, resource.id)
            state = resource.update(client, match[0].state)
        if state is not None:
            log.record(client, deployment.id, resource, state)
        else:
            print("deploy: nochange", type(resource).__name__, resource.id)
    for deployed in reversed(sort_deployed(deploy_state)):
        matching_resource = [e for e in all_resources if e.id == deployed.resource_id]
        if len(matching_resource) == 0:
            klas = globals()[deployed.resource_type]
            print("deploy: remove", deployed.resource_type, deployed.resource_id)
            klas.delete(client, deployed.state)
            log.remove(client, deployment.id, deployed.resource_id)

# called as a library
def deploy(deployment: Deployment, lusid_url: str, access_token: str):
    client = create_client(lusid_url, access_token)
    run(client, deployment)


def load_vars(vars_file):
    if vars_file:
        host_vars = SimpleNamespace(**json.load(vars_file))
    else:
        host_vars = SimpleNamespace()
    return host_vars
